HYD_RATE_KEYWORDS = ['QOIL', 'QLIQ', 'QGAS', 'QWAT', 'QWGAS']
HYD_GAS_RATIO_KEYWORDS = ['GOR', 'GLR', 'OGR', 'GWR', 'MMW']
HYD_WATER_RATIO_KEYWORDS = ['WCUT', 'WGR', 'OCUT', 'WWGR']
HYD_PRESSURE_KEYWORDS = ['PIN', 'POUT', 'THP']
HYD_ARRAY_KEYWORDS = HYD_RATE_KEYWORDS + HYD_GAS_RATIO_KEYWORDS + HYD_WATER_RATIO_KEYWORDS + HYD_GAS_RATIO_KEYWORDS + \
    HYD_PRESSURE_KEYWORDS
HYD_ALQ_KEYWORD = ['ALQ']
HYD_ALQ_OPTIONS = ['GASRATE', 'NOTGLIFT']

HYD_CORRECTION_KEYWORD = ['DATGRAD']
HYD_CORRECTION_OPTIONS = ['GRAD', 'OILGRAD', 'GASGRAD', 'WATERGRAD']

HYD_KEYWORDS_VALUE_FLOAT = ['LENGTH', 'DATUM', 'DZW']
HYD_SINGLE_KEYWORDS = ['NOCHK']

HYD_TABLE_HEADER_COLS = ['IGOR', 'IWCUT', 'IQOIL', 'IALG', 'BHP(ITHP)', 'IGLR', 'IQLIQ', 'IOGR', 'IWGR', 'IQGAS',
                         'IGWR', 'IOCUT', 'IQWAT', 'IMMW', 'IWWGR', 'IQWGAS']

HYD_LIMITS_KEYWORD = ['LIMITS']

HYD_WATINJ_KEYWORD = ['WATINJ']
HYD_WATINJ_KEYWORDS_VALUE_FLOAT = ['GRAD', 'VISC', 'DIAM', 'LENGTH', 'DZ', 'ROUGHNESS']

HYD_KEYWORDS = HYD_ARRAY_KEYWORDS + HYD_ALQ_KEYWORD + HYD_CORRECTION_KEYWORD + HYD_KEYWORDS_VALUE_FLOAT + \
    HYD_SINGLE_KEYWORDS + HYD_LIMITS_KEYWORD + HYD_WATINJ_KEYWORD
