# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

OPTIONS_KEYWORDS = ['ALL', 'AQUIFER', 'AUTO', 'CENTER', 'DATUM', 'DEF_DATUM', 'DESC', 'DP',
                    'ENDGLOBAL_METHOD_OVERRIDES', 'ENDGRIDTOPROC', 'ENDREGBLK', 'ENDREGDATA', 'ENDSUBREGIONS',
                    'EQ_LIST', 'FLEXDP', 'GLOBAL_METHOD_OVERRIDES', 'GRID', 'GRIDBLOCKS', 'GRIDTOPROC', 'INTSAT',
                    'NAME', 'NOCHK_SAL_TEMP', 'NONEQ', 'NUMBER', 'OFF', 'OVERREAD', 'PATTERN', 'PC', 'PD',
                    'POROSITY_INDEPENDENCE', 'PROCESS', 'PSTD', 'REGBLK', 'REGDATA', 'RESERVOIR', 'RES_TEMP', 'SOMOPT1',
                    'STONE1', 'SUBREGION', 'SUBREGIONS', 'SW', 'TSTD', 'VIP_INJ_PERF_KR_SCALING']
