# Currently ignoring functional form rel perm inputs, only tabular keywords are handled.
# Also currently ignoring spline options

RELPM_TABLE_KEYWORDS = ['WOTABLE', 'GOTABLE', 'GWTABLE', 'WOTABLE_IMB', 'GOTABLE_IMB', 'GWTABLE_IMB',
                        'WAGTABLE_WIMB', 'GRMTABLE_SGTR', 'GRMTABLE', 'GDWTABLE', 'LOW_SALTAB',
                        'PRSTAB', 'WAGTEST']
RELPM_SINGLE_KEYWORDS = ['KROINT', 'STONE2', 'STONE1', 'SOMOPT1', 'SOMOPT3', 'LOW_SAL', 'GW3PHASE',
                         'NOCHK', 'VIP_INJ_PERF_KR_SCALING', 'VEGO', 'VEGO_PC', 'VEWO', 'VEWO_PC',
                         'IRELPM_LOWIFT', 'KRWINT', 'STONE2_WAT', 'STONE1_WAT', 'SWMOPT1',
                         'FREEZE_PCGO', 'FREEZE_PCWO', 'FREEZE_PCGW', 'LOW_SAL_IRR', 'VIP_RELPM']
RELPM_HYSTERESIS_PRIMARY_KEYWORDS = ['NONE', 'KRW', 'KRG', 'KROW', 'PCWO', 'PCGO', 'WAG',
                                     'TOLREV', 'TOLHYS', 'NOCHK_HYS']
RELPM_KEYWORDS_VALUE_FLOAT = ['SOMOPT2', 'ST1EXP', 'ST1EPS', 'ST1NU', 'GRMDSW', 'PERMBASIS', 'PORBASIS', 'DRELPM',
                              'TENTHR', 'XEX', 'TENI', 'SWMOPT2']
RELPM_KEYWORDS_VALUE_STR = ['HYSTERESIS', 'JFUNC', 'SCALING', 'SCALING_PC', 'IFT', 'NEARCRIT', 'DERIVATIVES',
                            'RECONSTRUCT']
RELPM_NONDARCY_KEYWORDS = ['NONDARCY_GAS', 'NONDARCY_OIL']
RELPM_NONDARCY_PARAMS = ['BETA', 'BETA0', 'BETA1', 'BETA2', 'BETA3', 'BETA4', 'BETA5', 'IFT_THRES']

RELPM_KEYWORDS = RELPM_TABLE_KEYWORDS + RELPM_SINGLE_KEYWORDS + RELPM_KEYWORDS_VALUE_FLOAT + \
                 RELPM_KEYWORDS_VALUE_STR + RELPM_NONDARCY_KEYWORDS
