
ROCK_KEYWORDS_VALUE_FLOAT = ['CR', 'KP', 'PREF', 'TOLREV_P', 'TOLREV_SW', 'CR_REPRESSURE', 'KP_REPRESSURE',
                             'DPMIN', 'CREEPB', 'CREEPC', 'CREEPM']
ROCK_KEYWORDS_VALUE_STR = ['REVERSIBLE', 'IRREVERSIBLE']
ROCK_REV_IRREV_OPTIONS = ['ALL', 'CMTONLY', 'WIRCTONLY']
ROCK_SINGLE_KEYWORDS = ['COMPR', 'KPMULT', 'CREEP', 'TAMULTONLY']
ROCK_CMT = ['CMT']
ROCK_WATER_INDUCED_CMT = ['WIRCT', 'SWINIT']
ROCK_ALL_TABLE_KEYWORDS = ROCK_CMT + ROCK_WATER_INDUCED_CMT

# Apparent Permeability in Nano-porosity Rocks (Knudsen diffusion/slippage) options
ROCK_KNUDSEN_DIFF_KEYWORDS = ['JAVADPOUR', 'CIVAN', 'KNUDSEN']
ROCK_KNUDSEN_DIFF_OPTIONS = ['R', 'ALPHA', 'BETA', 'A', 'B']

ROCK_KEYWORDS = ROCK_KEYWORDS_VALUE_FLOAT + ROCK_KEYWORDS_VALUE_STR + ROCK_SINGLE_KEYWORDS + \
    ROCK_ALL_TABLE_KEYWORDS + ROCK_REV_IRREV_OPTIONS + ROCK_KNUDSEN_DIFF_KEYWORDS + ROCK_KNUDSEN_DIFF_OPTIONS
