FCS_KEYWORDS = ['RESERVOIR', 'EQUIL', 'EOS_DEFAULTS', 'TRACER_INIT', 'STRUCTURED_GRID', 'OPTIONS', 'ROCK', 'RELPM',
                'ADSORPTION', 'PVT', 'SEPARATOR', 'ALPHAF', 'POLYMER', 'WATER', 'AQUIFER', 'FLUXIN', 'RUNCONTROL',
                'WELLS', 'SURFACE', 'IPR', 'GASLIFT', 'VALVE', 'HYD', 'PUMP', 'COMPRESSOR', 'CHOKE', 'ICD', 'ESP']

# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

FCS_ALL_KEYWORDS = ['ADSORPTION', 'ALPHA', 'ALPHAF', 'AQUIFER', 'AQUIFER_FILES', 'CASE', 'CHOKE', 'COMPRESSOR',
                    'DATEFORMAT', 'DD/MM/YYYY', 'DEFAULT_SUNITS', 'DEFAULT_UNITS', 'DESC', 'ENGLISH', 'EOS_DEFAULTS',
                    'EQUIL', 'ESP', 'FIELD', 'FLUXIN', 'GASLIFT', 'GRID_FILES', 'HYD', 'ICD',
                    'INITIALIZATION_FILES', 'IPR', 'LGR', 'METBAR', 'METHOD', 'MM/DD/YYYY', 'NET_METHOD_FILES', 'ON',
                    'OPTIONS', 'OVERRIDE', 'POLYMER', 'PPM', 'PROPERTY_FILES', 'PUMP', 'PVT', 'PVT_FILES',
                    'RECURRENT_FILES', 'RELPM', 'RESERVOIR', 'RESTART', 'ROCK', 'ROCK_FILES', 'RUNCONTROL',
                    'RUN_SUNITS',
                    'RUN_UNITS', 'SEPARATOR', 'STRUCTURED_GRID', 'SURFACE', 'TIME', 'TRACER_INIT', 'VALVE', 'VIP',
                    'WATER',
                    'WELLS', 'X']
