# TODO: Missing time data related values - RESTART, PLUS

WELLS_KEYWORDS = ['ADJPERF', 'ANGLA', 'ANGLE', 'ANGLV', 'CELL', 'CON', 'D', 'DATEFORMAT', 'DBOT', 'DEPTH', 'DKH',
                  'DKRG', 'DKRW', 'DSGR', 'DSWR', 'DTOP', 'DZ', 'ENDADJPERF', 'ENGLISH', 'FLOWSECT', 'FM', 'GRID',
                  'GROUP', 'IPTN', 'IRELPM', 'IW', 'JW', 'K', 'KH', 'KHMIN', 'KHMULT', 'KRG_SGRO', 'KRG_SGRW',
                  'KRG_SGU', 'KRO_SGL', 'KRO_SGR', 'KRO_SWL', 'KRO_SWR', 'KRW_SGL', 'KRW_SGR', 'KRW_SWRO', 'KRW_SWU',
                  'L', 'LAYER', 'LENGTH', 'MD', 'MDCON', 'ND', 'PARENT', 'PORTYPE', 'PPERF', 'PVMIN', 'RADB', 'RADBP',
                  'RADW', 'RADWP', 'RKHMIN', 'RPVMIN', 'SECT', 'SGL', 'SGR', 'SGRO', 'SGRW', 'SGTR', 'SGU', 'SKIN',
                  'SOTR', 'STAT', 'SWL', 'SWR', 'SWRO', 'SWU', 'TEMP', 'TIME', 'VAR', 'WELLMOD', 'WELLSPEC', 'WI', 'X',
                  'Y', 'ZONE']
