# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

SURFACE_KEYWORDS = ['A', 'ACPGCR', 'ACTIONS', 'ACTIVATE', 'ACTIVE', 'ADD', 'ADDVALUE', 'ALL', 'ALL1D', 'ALLOW_BHMOVE',
                    'ANGLE', 'API', 'ARRAYBOUNDS', 'AVG', 'B', 'BACKFLOW', 'BEGGS', 'BHCON', 'BHDEPTH', 'BHP',
                    'BLACKOIL', 'C', 'CALC', 'CALCCOND', 'CALCCONS', 'CALCMETHOD', 'CALCMULT', 'CELLAVG', 'CELLGRAD',
                    'CLEAR', 'CLEARP', 'CLEARQ', 'COMPONENT', 'COMPONENTS', 'CON', 'CONCENTRATION', 'CONDEFAULTS',
                    'CONLIST', 'CONNECTION', 'CONSTRAINT', 'CONSTRAINTS', 'CONTEST', 'CONTROL', 'CONTYPE',
                    'CORRELATION', 'CROSSFLOW', 'CROSS_SHUT', 'CTRL', 'CTRLCOND', 'CTRLCONS', 'CTRLMETHOD', 'CWLIM',
                    'CYCLE', 'D', 'DAMP', 'DATA', 'DATGRAD', 'DATUM', 'DAY', 'DAYS', 'DDEPTH', 'DEACTIVATE', 'DEN_TEMP',
                    'DEPTH', 'DIAM', 'DO', 'DONTCLEAR', 'DP', 'DPADD', 'DPBHAVG', 'DPBHMX', 'DPERF', 'DT', 'DTMIN',
                    'DUNROS', 'E', 'ELEVPR', 'ELSE', 'ELSEIF', 'END', 'ENDACTIVATE', 'ENDCONDEFAULTS', 'ENDCONLIST',
                    'ENDCONSTRAINTS', 'ENDCONTEST', 'ENDDEACTIVATE', 'ENDDO', 'ENDELEVPR', 'ENDGASWELLS',
                    'ENDGUIDERATE', 'ENDIF', 'ENDNODECON', 'ENDNODELIST', 'ENDNODES', 'ENDPDCORR', 'ENDPROCS', 'ENDPT',
                    'ENDQMULT', 'ENDSKIP', 'ENDSTREAMS', 'ENDSTREAM_TRACER', 'ENDTARGET', 'ENDTEMPGR', 'ENDTEMPPR',
                    'ENDTGTCON', 'ENDWELLBORE', 'ENDWELLHEAD', 'ENDWELLLIST', 'ENDWELLS', 'EOS', 'ESP', 'EXIT', 'F',
                    'FLOW_UNSTABLE', 'FREQ', 'FRPGCR', 'GAS', 'GASLIFT', 'GASMOB', 'GASWATER', 'GASWELLS', 'GLIFT',
                    'GOR', 'GORLIM', 'GORPERF', 'GORQMULT', 'GORWELL', 'GRADCALC', 'GROUP_PERFS', 'GRPGCR', 'GUIDERATE',
                    'HAGEDORN', 'HAG_BEG', 'HEATTR', 'HTC', 'HYDRAULICS', 'HYDSTAB', 'IBAT', 'ID', 'IF', 'INCLUDE',
                    'INCREASE', 'INJMOB', 'INJTGTP', 'INJTGTPHC', 'INTEGER', 'INT_1D', 'IPVT', 'ITIME', 'IW', 'IWAT',
                    'K', 'LENGTH', 'LEVEL', 'LIMIT', 'LOGICAL', 'LOGICAL_1D', 'LOOP', 'LUMPED', 'MAX', 'MAXOPEN',
                    'MDIN', 'MDOUT', 'METHOD', 'MIXED', 'MOBGRAD', 'MONTH', 'MULT', 'MULTIFIELD', 'N', 'N1', 'N2',
                    'NAME', 'NETFILL', 'NETTEMP', 'NETVOID', 'NEWTON', 'NHC', 'NODECON', 'NODEIN', 'NODELIST',
                    'NODEOUT', 'NODES', 'NONE', 'NOONTIME', 'NOSLIP', 'NOTDATA', 'NOW', 'NUMBER', 'OFF', 'OIL',
                    'OILGRAD', 'ON', 'ONTIME', 'OUTPAVG', 'PARENT', 'PATTERN', 'PC', 'PD', 'PDCORR', 'PHASE', 'PIPE',
                    'PIPEGRAD', 'PLUS', 'PMAX', 'PMIN', 'PRIORITY', 'PROC', 'PROCABORT', 'PROCS', 'PRODTGTP',
                    'PRODTGTPHC', 'PRODWELLS', 'QALL', 'QALLMAX', 'QALLRMAX', 'QG', 'QGAS', 'QGLFT', 'QGNETPROD',
                    'QGSMAX', 'QGSMIN', 'QGUIDE', 'QLIQ', 'QLIQLFT', 'QLIQSMAX', 'QLIQSMIN', 'QMIN', 'QMIN_NOSHUT',
                    'QMULT', 'QMULTOPT', 'QO', 'QOIL', 'QOLFT', 'QOMAX', 'QOSMAX', 'QOSMIN', 'QW', 'QWAT', 'QWATER',
                    'QWLFT', 'QWMAX', 'QWSMAX', 'QWSMIN', 'RANKDT', 'RATE', 'REAL', 'REAL_1D', 'REGAVG', 'REGHCAVG',
                    'REGION', 'REMOVE', 'RESERVOIR', 'RFTWELLS', 'ROUGHNESS', 'SCALE', 'SEQUENTIAL', 'SETTING',
                    'SHCGAS', 'SHCOIL', 'SHCWAT', 'SHUTINON', 'SHUTIN_UNSTABLE', 'SINK', 'SKIP', 'SS_CONS', 'SS_NODES',
                    'SS_PERFS', 'SS_WELLS', 'START', 'STATIC', 'STATION', 'STD', 'STREAM', 'STREAMS', 'STREAM_TRACER',
                    'SURFACE', 'TARG', 'TARGET', 'TARGETS', 'TEMP', 'TEMPGR', 'TEMPPR', 'TEST', 'TGTCON', 'THEN',
                    'TIME', 'TRACER', 'TRACERS', 'TSTPRF', 'TVD', 'TYPE', 'VALUE', 'W', 'WATER', 'WATERINJWELLS',
                    'WCUT', 'WCUTMAX', 'WCUTPERF', 'WCUTPLUG', 'WELL', 'WELLBORE', 'WELLCONTROL', 'WELLDATUM',
                    'WELLGRAD', 'WELLHEAD', 'WELLLIST', 'WELLS', 'WI', 'WIMULT', 'X', 'YEAR']

CONNECTION_KEYWORDS = ['NAME', 'NODEIN', 'NODEOUT', 'IPVT', 'IWAT', 'IBAT', 'TYPE', 'METHOD', 'LENGTH', 'NUMBER',
                       'DDEPTH', 'DIAM', 'INNERDIAM', 'ROUGHNESS', 'ELEVPR', 'TEMPPR', 'MDIN', 'MDOUT', 'HTC', 'DPADD',
                       'RATEMULT', 'TEMPIN', 'TEMPOUT', 'TRACERS', 'POLYMER', 'DTADD', 'SEAWPR', 'THICKNESS', 'INSTHN',
                       'INSK', 'GRPGCR', 'FRPGCR', 'ACPGCR',
                       ]
