# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

RUNCONTROL_KEYWORDS = ['2PHASE', 'ADJUSTTOTIME', 'ALL', 'ALLOW', 'AQFLUX', 'ARRAYS', 'AUTO', 'AVGSG', 'AVGSO', 'AVGSW',
                       'BG', 'BHP', 'BOF', 'CGFLUX', 'CGI', 'CGLG', 'CGP', 'CLP', 'COFLUX', 'COMP', 'CONCENTRATION',
                       'CONNECTFILE', 'COP', 'CPR_PRESSURE_EQUATION', 'CRGI', 'CRGP', 'CROP', 'CRWI', 'CRWP', 'CWFLUX',
                       'CWI', 'CWP', 'DATE', 'DATEFORMAT', 'DCMAX', 'DD/MM/YYYY', 'DEBUG', 'DEN', 'DEPTHTAB',
                       'DISALLOW', 'DRDN', 'DRMX', 'DRSDT', 'DSGLIM', 'DSLIM', 'DT', 'DUAL_SOLVER', 'DVOLLIM', 'DZLIM',
                       'END', 'ENDDEBUG', 'ENDGRIDS', 'ENDMAPOUT', 'ENDOUTPUT', 'ENDOUTSTART', 'ENDPLOTOUT',
                       'ENDPOINTS', 'ENDREGIONOUT', 'ENDSPREADSHEET', 'ENDSSOUT', 'EQUILREGION', 'EXTENDED',
                       'FACILITIES', 'FIELD', 'FIELDPLOT', 'FREEGAS', 'FREQ', 'FULLMAPARRAYS', 'GAS', 'GLOBALTOL',
                       'GOR', 'GRIDS', 'GRIDSOLVER', 'HYDMAX', 'HYDRAULICS', 'ILU', 'IMPES', 'IMPES_COUPLING',
                       'IMPLICIT', 'IMPLICITMBAL', 'IMPLICIT_COUPLING', 'IMPLICIT_PRECON', 'IMPSTAB', 'INPLACE',
                       'ITERATIONS', 'KMCMT', 'KR', 'LIMIT', 'LOGOUT', 'MAPBINARY', 'MAPFORM', 'MAPOUT', 'MAPS',
                       'MAPVDB', 'MASS', 'MAX', 'MAXBADNETS', 'MAXINCREASE', 'MAXNEWTONS', 'METHOD', 'MIN', 'MOB',
                       'MONTHLY', 'NAME', 'NETPLOT', 'NETSUM', 'NETWORK', 'NOCPUOUT', 'NOCUT', 'NONE', 'NORECOUT',
                       'NSC', 'OFF', 'OIL', 'ON', 'ORDER2009', 'OREC', 'OUTPUT', 'OUTSTART', 'OVERSHOOT', 'P', 'PAVH',
                       'PAVT', 'PC', 'PCELL', 'PDATUMHC', 'PDATUMT', 'PDCOR', 'PERF', 'PERFPLOT', 'PERFREV', 'PERFS',
                       'PERFTRACER', 'PI', 'PIN', 'PLOTBINARY', 'PLOTFORM', 'PLOTOUT', 'PLOTVDB', 'PMAX', 'PMIN',
                       'PNODE', 'POUT', 'PSAT', 'PV', 'PVMLT', 'QGI', 'QGLG', 'QGP', 'QLP', 'QMAXINJ', 'QMAXPROD',
                       'QOP', 'QUARTERLY', 'QWI', 'QWP', 'REGIONOUT', 'REGIONPLOT', 'REGIONS', 'RESERVOIR', 'RESTART',
                       'RFT', 'RFTFILE', 'RGI', 'RGP', 'ROCK', 'ROP', 'RS', 'RVG', 'RVO', 'RVPVH', 'RVPVT', 'RVW',
                       'RWI', 'RWP', 'SAL', 'SAT', 'SEPARATOR', 'SOLVER', 'SPREADSHEET', 'SSCSV', 'SSNOSORT', 'SSOUT',
                       'SSTAB', 'START', 'STAT', 'STGIP', 'STOIP', 'STOP', 'STREAMCALC', 'STWIP', 'SW', 'TARGETPLOT',
                       'TARGETS', 'TC', 'THP', 'TIME', 'TIMES', 'TNEXT', 'TOLS', 'TOTAL', 'TRANS', 'TSNUM', 'VIPUNITS',
                       'VISC', 'VOLCON', 'WATER', 'WCUT', 'WELL', 'WELLPLOT', 'WELLS', 'WOR', 'WPAVE', 'WPPV', 'WPWBC',
                       'X', 'Y', 'YEARLY', 'Z']
